@@ -7,20 +7,69 @@
 """Extractors for https://bato.to/"""
 
 from .common import Extractor, ChapterExtractor, MangaExtractor
-from .. import text, exception
-import re
+from .. import text, util
 
-BASE_PATTERN = (r"(?:https?://)?(?:"
-                r"(?:ba|d|h|m|w)to\.to|"
+BASE_PATTERN = (r"(?:https?://)?("
+                r"(?:ba|d|f|h|j|m|w)to\.to|"
                 r"(?:(?:manga|read)toto|batocomic|[xz]bato)\.(?:com|net|org)|"
                 r"comiko\.(?:net|org)|"
                 r"bat(?:otoo|o?two)\.com)")
 
+#  https://rentry.co/batoto
+DOMAINS = {
+    "dto.to",
+    "fto.to",
+    "hto.to",
+    "jto.to",
+    "mto.to",
+    "wto.to",
+    "xbato.com",
+    "xbato.net",
+    "xbato.org",
+    "zbato.com",
+    "zbato.net",
+    "zbato.org",
+    "readtoto.com",
+    "readtoto.net",
+    "readtoto.org",
+    "batocomic.com",
+    "batocomic.net",
+    "batocomic.org",
+    "batotoo.com",
+    "batotwo.com",
+    "comiko.net",
+    "comiko.org",
+    "battwo.com",
+}
+LEGACY_DOMAINS = {
+    "bato.to",
+    "mangatoto.com",
+    "mangatoto.net",
+    "mangatoto.org",
+}
+
 
 class BatotoBase():
     """Base class for batoto extractors"""
     category = "batoto"
-    root = "https://bato.to"
+    root = "https://xbato.org"
+    _warn_legacy = True
+
+    def _init_root(self):
+        domain = self.config("domain")
+        if domain is None or domain in {"auto", "url"}:
+            domain = self.groups[0]
+            if domain in LEGACY_DOMAINS:
+                if self._warn_legacy:
+                    BatotoBase._warn_legacy = False
+                    self.log.warning("Legacy domain '%s'", domain)
+        elif domain == "nolegacy":
+            domain = self.groups[0]
+            if domain in LEGACY_DOMAINS:
+                domain = "xbato.org"
+        elif domain == "nowarn":
+            domain = self.groups[0]
+        self.root = "https://" + domain
 
     def request(self, url, **kwargs):
         kwargs["encoding"] = "utf-8"
@@ -28,15 +77,16 @@ class BatotoBase():
 
 
 class BatotoChapterExtractor(BatotoBase, ChapterExtractor):
-    """Extractor for bato.to manga chapters"""
+    """Extractor for batoto manga chapters"""
+    archive_fmt = "{chapter_id}_{page}"
     pattern = BASE_PATTERN + r"/(?:title/[^/?#]+|chapter)/(\d+)"
-    example = "https://bato.to/title/12345-MANGA/54321"
+    example = "https://xbato.org/title/12345-MANGA/54321"
 
     def __init__(self, match):
-        self.root = text.root_from_url(match.group(0))
-        self.chapter_id = match.group(1)
-        url = "{}/title/0/{}".format(self.root, self.chapter_id)
-        ChapterExtractor.__init__(self, match, url)
+        ChapterExtractor.__init__(self, match, False)
+        self._init_root()
+        self.chapter_id = self.groups[1]
+        self.page_url = f"{self.root}/title/0/{self.chapter_id}"
 
     def metadata(self, page):
         extr = text.extract_from(page)
@@ -53,9 +103,9 @@ class BatotoChapterExtractor(BatotoBase, ChapterExtractor):
             info = text.remove_html(extr('link-hover">', "</"))
         info = text.unescape(info)
 
-        match = re.match(
+        match = util.re(
             r"(?i)(?:(?:Volume|S(?:eason)?)\s*(\d+)\s+)?"
-            r"(?:Chapter|Episode)\s*(\d+)([\w.]*)", info)
+            r"(?:Chapter|Episode)\s*(\d+)([\w.]*)").match(info)
         if match:
             volume, chapter, minor = match.groups()
         else:
@@ -86,25 +136,25 @@ class BatotoChapterExtractor(BatotoBase, ChapterExtractor):
 
 
 class BatotoMangaExtractor(BatotoBase, MangaExtractor):
-    """Extractor for bato.to manga"""
+    """Extractor for batoto manga"""
     reverse = False
     chapterclass = BatotoChapterExtractor
     pattern = (BASE_PATTERN +
                r"/(?:title/(\d+)[^/?#]*|series/(\d+)(?:/[^/?#]*)?)/?$")
-    example = "https://bato.to/title/12345-MANGA/"
+    example = "https://xbato.org/title/12345-MANGA/"
 
     def __init__(self, match):
-        self.root = text.root_from_url(match.group(0))
-        self.manga_id = match.group(1) or match.group(2)
-        url = "{}/title/{}".format(self.root, self.manga_id)
-        MangaExtractor.__init__(self, match, url)
+        MangaExtractor.__init__(self, match, False)
+        self._init_root()
+        self.manga_id = self.groups[1] or self.groups[2]
+        self.page_url = f"{self.root}/title/{self.manga_id}"
 
     def chapters(self, page):
         extr = text.extract_from(page)
 
-        warning = extr(' class="alert alert-warning">', "</div><")
+        warning = extr(' class="alert alert-warning">', "</div>")
         if warning:
-            raise exception.StopExtraction("'%s'", text.remove_html(warning))
+            self.log.warning("'%s'", text.remove_html(warning))
 
         data = {
             "manga_id": text.parse_int(self.manga_id),
@@ -127,6 +177,6 @@ class BatotoMangaExtractor(BatotoBase, MangaExtractor):
             data["date"] = text.parse_datetime(
                 extr('time="', '"'), "%Y-%m-%dT%H:%M:%S.%fZ")
 
-            url = "{}/title/{}".format(self.root, href)
+            url = f"{self.root}/title/{href}"
             results.append((url, data.copy()))
         return results
